"""Mixin classes for unittests"""
import contextlib
import io


class AssertPrintsMixin:
  """Mixin class for print assertion"""
  @contextlib.contextmanager
  def assert_prints(self, target: str):
    """Assert that contextual code prints the target string.
    Use as context manager

    Args:
      target (str): Print expectation"""
    with io.StringIO() as buf, contextlib.redirect_stdout(buf):
      yield
      printed = buf.getvalue()
    self.assertEqual(printed, target)



"""Generate badge URLs from reports

Example:

.. code:: bash

  python -m coverage run -m unittest
  python -m coverage json --omit "*/tests/*,*/venv/*"
  python -m chromatictools.badges coverage"""
from chromatictools import colors, cli
import argparse
import json
from typing import Union
import sys


class HuePercentBadge:
  """Badge URL generator class based on mapping percentage to hue

  Args:
    filename (str): File to read for informtion
    label (str): Badge label
    hue_0 (float): Hue at 0 percent. Default is 0 (red)
    hue_100 (float): Hue at 100 percent. Default is 130 (green)
    hue_gamma (float): Gamma exponent for hue scale correction
    saturation (float): Color saturation
    value (float): Color value
    fmt (str): URL format string with placeholders for :data:`label`,
      :data:`message` and :data:`color`"""
  def __init__(
    self,
    filename: str,
    label: str,
    hue_0: float = 0,
    hue_100: float = 130,
    hue_gamma: float = 1,
    saturation: float = 1,
    value: float = 1,
    fmt: str = "https://img.shields.io/badge/{label}-{message}-{color}",
  ):
    self.filename = filename
    self.label = label
    self.hue_0 = hue_0
    self.hue_100 = hue_100
    self.hue_gamma = hue_gamma
    self.saturation = saturation
    self.value = value
    self.fmt = fmt

  @property
  def percent(self) -> Union[str, float]:
    """Percentage score"""
    return "N/A"

  @property
  def message(self) -> str:
    """Badge message"""
    p = self.percent
    if not isinstance(p, str):
      p = "{:.0f}%25".format(p)
    return p

  @property
  def hue(self) -> float:
    """Color hue for the badge"""
    p = self.percent
    if isinstance(p, str):
      return self.hue_0
    p = (p / 100) ** self.hue_gamma
    return p * self.hue_100 + (1 - p) * self.hue_0

  @property
  def color(self) -> str:
    """Hexadecimal color string for the badge"""
    return colors.rgb2hex(
      *colors.hsv2rgb(
        self.hue,
        self.saturation,
        self.value
      )
    )

  @property
  def url(self) -> str:
    """Badge URL string"""
    return self.fmt.format(
      label=self.label,
      message=self.message,
      color=self.color,
    )

  def __str__(self) -> str:
    """Get badge URL"""
    return self.url


class CoverageBadge(HuePercentBadge):
  """Coverage badge utility class

  Args:
    filename (str): Coverage report JSON file
    label (str): Badge label. Default is :data:`"coverage"`
    hue_gamma (float): Gamma exponent for hue scale correction.
      Default is :data:`6`
    kwargs: Keyword arguments for :class:`HuePercentBadge`"""
  def __init__(
    self,
    filename: str,
    label: str = "coverage",
    hue_gamma: float = 6,
    **kwargs,
  ):
    super().__init__(
      filename=filename,
      label=label,
      hue_gamma=hue_gamma,
      **kwargs
    )

  @property
  def report(self) -> dict:
    """Coverage report dictionary"""
    try:
      with open(self.filename, "r") as f:
        d = json.load(f)
    except FileNotFoundError:
      d = {}
    return d

  @property
  def percent(self) -> Union[str, float]:
    """Coverage percentage if found in file, else :data:`"N/A"`"""
    p = self.report.get("totals", {}).get("percent_covered", None)
    if p is None:
      p = super().percent
    return p


class PylintBadge(HuePercentBadge):
  """Pylint badge utility class

  Args:
    filename (str): Text file output of pylint
    label (str): Badge label. Default is :data:`"pylint"`
    kwargs: Keyword arguments for :class:`HuePercentBadge`"""
  def __init__(
    self,
    filename: str,
    label: str = "pylint",
    **kwargs,
  ):
    super().__init__(
      filename=filename,
      label=label,
      **kwargs
    )
    self.filename = filename

  @property
  def report(self) -> str:
    """Pylint report text file"""
    try:
      with open(self.filename, "r") as f:
        s = f.read()
    except FileNotFoundError:
      s = ""
    return s

  @property
  def percent(self) -> Union[str, float]:
    """Pylint rating (out of 100) if found in file, else :data:`"N/A"`"""
    s = self.report
    prefix = "Your code has been rated at "
    if prefix not in s:
      return super().percent
    return float(s.split(prefix, 1)[-1].split("/", 1)[0]) * 10

  @property
  def message(self) -> str:
    """Badge Message"""
    p = self.percent
    if not isinstance(p, str):
      p = "{:.2f}%2F10".format(p / 10)
    return p


@cli.main(__name__, *sys.argv[1:])
def main(*argv):
  """CLI launcher function

  Args:
    argv: Command-line arguments

  Returns:
    int: Exit code"""
  cmds = {
    "coverage": lambda a: CoverageBadge(a.coverage),
    "pylint": lambda a: PylintBadge(a.pylint),
  }
  parser = argparse.ArgumentParser(description=globals()["__doc__"])
  parser.add_argument(
    "badge",
    help="Which badge to use. Should be in {}".format(list(cmds.keys()))
  )
  parser.add_argument(
    "-c",
    metavar="filename",
    dest="coverage",
    default="coverage.json",
    required=False,
    help="JSON coverage report filepath",
  )
  parser.add_argument(
    "-l",
    metavar="filename",
    dest="pylint",
    default="pylint.txt",
    required=False,
    help="Pylint report filepath",
  )
  args = parser.parse_args(argv)
  print(cmds[args.badge](args))
  return 0



"""Utilities for pickle"""
import functools
import pickle
import os
from typing import Any, Callable, Union


def read_pickled(filepath: str, encoding: str = "latin1") -> Any:
  """Read the pickled data in the file

    Args:
      filepath (str): The file path
      encoding (str): File encoding. Defaults to :data:`"latin1"`

    Returns:
      The unpickled data"""
  with open(filepath, "rb") as f:
    return pickle.load(f, encoding=encoding)


def save_pickled(obj: Any, filepath: str):
  """Save data in a pickled file

    Args:
      obj (Any): The data to pickle
      filepath (str): The path of the file to write.
        Directory is created if it doesn't exist"""
  d = os.path.dirname(filepath)
  if d:
    os.makedirs(d, exist_ok=True)
  with open(filepath, "wb") as f:
    pickle.dump(obj, f)


def _pickled_cache_m(
  filepath: Callable[[Any], str]
) -> Callable[[Callable], Callable]:
  """Store the last result of the function call
  in a pickled file (method version). See
  :func:`pickled_cache` for usage example

  Args:
    filepath (Callable): A method that takes no other
      arguments other than the instance and returns the
      path of the file to read/write

  Returns:
    Callable[[Callable], Callable]: function decorator.
    The decorated function will also have an attribute
    function 'forced', that calls the function forcing
    cache overwriting"""
  def pickled_cache_(foo: Callable) -> Callable:
    @functools.wraps(foo)
    def foo_forced(*args, **kwargs):
      r = foo(*args, **kwargs)
      save_pickled(r, filepath(*args[:1]))
      return r

    @functools.wraps(foo)
    def foo_cached(*args, **kwargs):
      return read_pickled(filepath(*args[:1]))\
        if os.path.isfile(filepath(*args[:1]))\
        else foo_forced(*args, **kwargs)

    foo_cached.forced = foo_forced
    return foo_cached
  return pickled_cache_


def _pickled_cache_s(filepath: str) -> Callable[[Callable], Callable]:
  """Store the last result of the function call
  in a pickled file (string version)

  Args:
    filepath (str): The path of the file to read/write

  Returns:
    Callable[[Callable], Callable]: function decorator.
    The decorated function will also have an attribute
    function 'forced', that calls the function forcing
    cache overwriting"""
  return _pickled_cache_m(lambda *args, **kwargs: filepath)


def _pickled_cache_p(
  filepath: property
) -> Callable[[Callable], Callable]:
  """Store the last result of the function call
  in a pickled file (property version). See
  :func:`pickled_cache` for an usage example

  Args:
    filepath (property): A property that returns the
      path of the file to read/write

  Returns:
    Callable[[Callable], Callable]: function decorator.
    The decorated function will also have an attribute
    function 'forced', that calls the function forcing
    cache overwriting"""
  return _pickled_cache_m(filepath.fget)


_pickled_cache_funcs = (
  (str, _pickled_cache_s),
  (property, _pickled_cache_p),
  (Callable, _pickled_cache_m),
)


def pickled_cache(
  filepath: Union[str, Callable, property]
) -> Callable[[Callable], Callable]:

  """Store the last result of the function call
  in a pickled file (string version)

  Args:
    filepath (str or property or method): The path of the file to read/write as
     a string or as a getter method or property

  Returns:
    Callable[[Callable], Callable]: function decorator.
    The decorated function will also have an attribute
    function 'forced', that calls the function forcing
    cache overwriting

  Example:
    >>> from chromatictools.pickle import pickled_cache
    >>> cache_file: str = ".pickled_example.cache"
    >>> @pickled_cache(cache_file)
    ... def echo(s: str):
    ...   return s
    >>> echo("Alice")
    'Alice'
    >>> echo("Bob")
    'Alice'
    >>> echo.forced("Bob")
    'Bob'
    >>> echo("Alice")
    'Bob'
    >>> import os
    >>> os.remove(cache_file)

  Remove cache for inter-session repeatability (otherwise,
  the second time the example is run, the first `echo("Alice")`
  would return `"Bob"`). In a real scenario, you probably wouldn't
  want to do this, as inter-session memory is the main purpose
  of this decorator
  """
  for c, f in _pickled_cache_funcs:
    if isinstance(filepath, c):
      return f(filepath)
  raise NotImplementedError(
    "No 'pickled_cache' decorator available for filepath of type '{}'".format(
      type(filepath).__name__
    )
  )



"""Utilities for color handling"""
from typing import Tuple


def hsv2rgb(h: float , s: float, v: float) -> Tuple[int, int, int]:
  """Convert HSV to RGB

  Args:
    h (float): Hue (degrees)
    s (float): Saturation (between 0 and 1)
    v (float): Value (between 0 and 1)

  Returns
    tuple of int: RGB values (between 0 and 255)"""
  h = h % 360
  c = v * s
  x = c * (1 - abs((h / 60) % 2 - 1))
  m = v - c
  t = (
    (c, x, 0),
    (x, c, 0),
    (0, c, x),
    (0, x, c),
    (x, 0, c),
    (c, 0, x),
  )[int(h) // 60]
  return tuple(
    int((i + m) * 255)
    for i in t
  )


def rgb2hex(r: int, g: int, b: int) -> str:
  """Convert RGB integers to hexadecimal string

  Args:
    r (int): Red (between 0 and 255)
    g (int): Green (between 0 and 255)
    b (int): Blue (between 0 and 255)

  Returns
    str: Hexadecimal string"""
  return "".join(
    hex(v)[2:].rjust(2, "0")
    for v in (r, g, b)
  )



"""Utilities for extracting unittests from doctests"""
import unittest
import doctest
import contextlib
import io
from typing import Tuple


class DocTestMeta(type):
  """Metaclass for generating :class:`unittest.TestCase` classes.
  Modules to test should be listed in an attribute called :data:`_modules`"""
  def __new__(mcs: type, name: str, bases: Tuple[type, ...], dct: dict):
    if len(bases) == 0:
      bases = (unittest.TestCase,)
    for m in dct.get("_modules", ()):
      def method(self, mod=m):
        with io.StringIO() as buf, contextlib.redirect_stdout(buf):
          res = doctest.testmod(mod, verbose=True)
          msg = buf.getvalue()
        self.assertFalse(bool(res.failed), msg=msg)
      method.__name__ = "test_" + m.__name__.replace(".", "_")
      method.__doc__ = m.__doc__.split("\n", 1)[0]
      dct[method.__name__] = method
    cls = super().__new__(mcs, name, bases, dct)
    return cls



"""Utilities for Command-Line Interface applications"""
import sys


def main(name: str, *args, **kwargs):
  """Decorator for "main" functions of modules

  Args:
    name (str): Module name. Builtin variable :data:`__name__` should be
      passed as argument
    args: Positional arguments for execution as main function
    kwargs: Keyword arguments for execution as main function"""
  def main_(func):
    if name == "__main__":
      sys.exit(func(*args, **kwargs))
    return func
  return main_



"""Utilities for imports"""
import importlib
import types
from typing import Union


def import_if_installed(
  name: str
) -> Union[types.ModuleType, ModuleNotFoundError]:
  """Import a module if it is installed. Otherwise return a
  :class:`ModuleNotFoundError`

  Args:
    name (str): Module name

  Returns:
    module or ModuleNotFoundError: The module, if it is found. Otherwise a
    :class:`ModuleNotFoundError` instance"""
  try:
    module = importlib.import_module(name)
  except ModuleNotFoundError as e:
    return e
  else:
    return module



import setuptools
from chromatictools import __version__ as version


with open("README.md", "r") as f:
  readme = f.read()


setuptools.setup(
  name="chromatictools",
  version=version,
  author="Marco Tiraboschi",
  author_email="marco.tiraboschi@unimi.it",
  maintainer="Marco Tiraboschi",
  maintainer_email="marco.tiraboschi@unimi.it",
  description="chromatictools",
  long_description=readme,
  long_description_content_type="text/markdown",
  url="https://github.com/ChromaticIsobar/chromatictools",
  packages=setuptools.find_packages(
    include=["chromatictools", "chromatictools.*"]
  ),
  include_package_data=True,
  setup_requires=[
    "wheel",
  ],
  install_requires=[
    "gin-config",
    "numpy",
  ],
  extras_require={
    "notebooks": [
      "jupyter",
      "matplotlib",
      "seaborn",
      "tqdm",
    ],
    "docs": [
      "sphinx",
      "sphinx_rtd_theme",
      "m2r2",
      "recommonmark",
    ],
    "lint": [
      "pylint",
    ],
    "cov": [
      "coverage",
    ],
  },
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
)

